@@ -36,7 +36,7 @@ where
 /// This simplifies implementing decoders that do not have any special handling
 /// for length-prefixed types.
 #[doc(hidden)]
-pub struct RemainingSimpleDecoder<R, I, L>
+pub struct LimitedStorageDecoder<R, I, L>
 where
     I: IntegerEncoding,
     L: UsizeEncoding,
@@ -52,13 +52,14 @@ where
     L: UsizeEncoding,
 {
     type Error = R::Error;
-    type Pack = Self;
+    type Pack<'this> = Self;
     type Some = Self;
-    type Sequence = RemainingSimpleDecoder<R, I, L>;
-    type Map = RemainingSimpleDecoder<R, I, L>;
-    type Struct = RemainingSimpleDecoder<R, I, L>;
-    type Tuple = RemainingSimpleDecoder<R, I, L>;
-    type Variant = Self;
+    type Sequence<'this> = LimitedStorageDecoder<R, I, L>;
+    type Tuple<'this> = Self;
+    type Map<'this> = LimitedStorageDecoder<R, I, L>;
+    type Struct<'this> = LimitedStorageDecoder<R, I, L>;
+    type TupleStruct<'this> = LimitedStorageDecoder<R, I, L>;
+    type Variant<'this> = Self;
 
     #[inline]
     fn expecting(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
@@ -81,8 +82,11 @@ where
     }
 
     #[inline]
-    fn decode_pack(self) -> Result<Self::Pack, Self::Error> {
-        Ok(self)
+    fn decode_pack<T, O>(self, decoder: T) -> Result<O, Self::Error>
+    where
+        T: FnOnce(Self::Pack<'_>) -> Result<O, Self::Error>,
+    {
+        decoder(self)
     }
 
     #[inline]
@@ -241,23 +245,43 @@ where
     }
 
     #[inline]
-    fn decode_sequence(self) -> Result<Self::Sequence, Self::Error> {
-        RemainingSimpleDecoder::new(self)
+    fn decode_sequence<T, O>(self, decoder: T) -> Result<O, Self::Error>
+    where
+        T: FnOnce(Self::Sequence<'_>) -> Result<O, Self::Error>,
+    {
+        decoder(LimitedStorageDecoder::new(self)?)
     }
 
     #[inline]
-    fn decode_map(self) -> Result<Self::Map, Self::Error> {
-        RemainingSimpleDecoder::new(self)
+    fn decode_tuple<T, O>(self, _: usize, decoder: T) -> Result<O, Self::Error>
+    where
+        T: FnOnce(Self::Tuple<'_>) -> Result<O, Self::Error>,
+    {
+        decoder(self)
     }
 
     #[inline]
-    fn decode_struct(self, _: usize) -> Result<Self::Struct, Self::Error> {
-        RemainingSimpleDecoder::new(self)
+    fn decode_map<T, O>(self, decoder: T) -> Result<O, Self::Error>
+    where
+        T: FnOnce(Self::Map<'_>) -> Result<O, Self::Error>,
+    {
+        decoder(LimitedStorageDecoder::new(self)?)
     }
 
     #[inline]
-    fn decode_tuple(self, _: usize) -> Result<Self::Tuple, Self::Error> {
-        RemainingSimpleDecoder::new(self)
+    fn decode_struct<T, O>(self, _: usize, decoder: T) -> Result<O, Self::Error>
+    where
+        T: FnOnce(Self::Struct<'_>) -> Result<O, Self::Error>,
+    {
+        decoder(LimitedStorageDecoder::new(self)?)
+    }
+
+    #[inline]
+    fn decode_tuple_struct<T, O>(self, _: usize, decoder: T) -> Result<O, Self::Error>
+    where
+        T: FnOnce(Self::TupleStruct<'_>) -> Result<O, Self::Error>,
+    {
+        decoder(LimitedStorageDecoder::new(self)?)
     }
 
     #[inline]
@@ -266,8 +290,11 @@ where
     }
 
     #[inline]
-    fn decode_variant(self) -> Result<Self::Variant, Self::Error> {
-        Ok(self)
+    fn decode_variant<T, O>(self, decoder: T) -> Result<O, Self::Error>
+    where
+        T: FnOnce(Self::Variant<'_>) -> Result<O, Self::Error>,
+    {
+        decoder(self)
     }
 }
 
@@ -284,14 +311,9 @@ where
     fn next(&mut self) -> Result<Self::Decoder<'_>, Self::Error> {
         Ok(StorageDecoder::new(&mut self.reader))
     }
-
-    #[inline]
-    fn end(self) -> Result<(), Self::Error> {
-        Ok(())
-    }
 }
 
-impl<'de, R, I, L> RemainingSimpleDecoder<R, I, L>
+impl<'de, R, I, L> LimitedStorageDecoder<R, I, L>
 where
     R: PositionedReader<'de>,
     I: IntegerEncoding,
@@ -304,7 +326,7 @@ where
     }
 }
 
-impl<'de, R, I, L> SequenceDecoder<'de> for RemainingSimpleDecoder<R, I, L>
+impl<'de, R, I, L> SequenceDecoder<'de> for LimitedStorageDecoder<R, I, L>
 where
     R: PositionedReader<'de>,
     I: IntegerEncoding,
@@ -329,7 +351,7 @@ where
     }
 }
 
-impl<'de, R, I, L> PairsDecoder<'de> for RemainingSimpleDecoder<R, I, L>
+impl<'de, R, I, L> PairsDecoder<'de> for LimitedStorageDecoder<R, I, L>
 where
     R: PositionedReader<'de>,
     I: IntegerEncoding,
@@ -365,7 +387,7 @@ where
 {
     type Error = R::Error;
     type First<'this> = StorageDecoder<&'this mut R, I, L> where Self: 'this;
-    type Second = StorageDecoder<R, I, L>;
+    type Second<'this> = StorageDecoder<&'this mut R, I, L> where Self: 'this;
 
     #[inline]
     fn first(&mut self) -> Result<Self::First<'_>, Self::Error> {
@@ -373,12 +395,12 @@ where
     }
 
     #[inline]
-    fn second(self) -> Result<Self::Second, Self::Error> {
-        Ok(StorageDecoder::new(self.reader))
+    fn second(&mut self) -> Result<Self::Second<'_>, Self::Error> {
+        Ok(StorageDecoder::new(&mut self.reader))
     }
 
     #[inline]
-    fn skip_second(self) -> Result<bool, Self::Error> {
+    fn skip_second(&mut self) -> Result<bool, Self::Error> {
         Ok(false)
     }
 }
